import 'app_clock_service/app_clock.dart';
import 'background_fetch/background_fetch.dart';
import 'configuration/configuration_service.dart';
import 'data_classes/exposure_data.dart';
import 'data_service/data_service.dart';
import 'detection_service/detection_service.dart';
import 'gps_service/gps_service.dart';
import 'log/log.dart';
import 'nearby_connections_service/nearby_connections_proximity_detection.dart';
import 'registration/registration.dart';

class Heroes
{
  static final _instance = Heroes._internal();
  /// Heroes is a singleton.
  factory Heroes() => _instance;
  Heroes._internal() {
    _configuration = ConfigurationService();
    _registration = Registration();
    _dataService = DataService();
    _gpsService = GPSService();
    _detectionService = DetectionService();
    _backgroundFetchService = BackgroundFetchService();
  }

  /// Callback function that will be invoked when [ExposureData] changes.
  Function(ExposureData) onExposureDetected;

  ConfigurationService _configuration;
  Registration _registration;
  DataService _dataService;
  GPSService _gpsService;
  DetectionService _detectionService;
  BackgroundFetchService _backgroundFetchService;

  /// Returns whether or not the user is registered.
  Future<bool> get userIsRegistered async => await _registration.isRegistered;

  /// Registers a user.
  /// Returns whether or not registration was successful.
  Future<bool> registerUser(String invitationCode) async
  {
    if (invitationCode == null || invitationCode == "")
      return false;

    final result = await _registration.registerUser(inviteCode: invitationCode);

    if (result == null)
      return false;

    if (result.success)
      print("HEROES_FLUTTER_PLUGIN :: Successfully registered user.");
    else
      print("HEROES_FLUTTER_PLUGIN :: User was not registerd.");

    return result.success;
  }

  String _apiKey = "";

  /// Sets API key and authenticates your application with the server.
  /// Returns whether or not authentication was successful.
  Future<bool> setAPIkey(String value) async
  {
    _apiKey = value;

    bool success = false;

    if (!(_configuration.isLoaded))
      success = await _configuration.loadConfiguration(apiKey: _apiKey);

    return success;
  }

  /// The device user's ID.
  String get userID => _registration.userID;

  /// The invitation codes the user got after registering. They are single-use.
  List<String> get invitationCodes => _registration.invitationCodes;

  /// Whether or not Heroes is started.
  bool get isStarted => _isStarted;
  bool _isStarted = false;

  bool _isStarting = false;
  bool _isStopping = false;

  /// Starts detection of other devices and its handling.
  /// Returns whether or not it was successful. Will return false if Heroes is currently stopping, the user isn't registered, or a valid API key hasn't been set.
  Future<bool> start({String APIkey = ""}) async
  {
    if (_isStopping)
      return false;

    if (_isStarted || _isStarting)
      return true;

    if (APIkey != null && APIkey != "")
      _apiKey = APIkey;
    else if (_apiKey == "")
      return false;

    _isStarting = true;

    Log.addToLog("Starting Heroes with API key: $_apiKey");

    if (!(_configuration.isLoaded))
      await _configuration.loadConfiguration(apiKey: _apiKey);

    AppClock(); //Initialize AppClock's static values.

    if (!(await userIsRegistered)) {
      Log.addToLog("User is not registered.");
      return false;
    }
    else
      Log.addToLog("User is registered. ID = " + _registration.userID.substring(0, 6) + ".");

    if (!(_dataService.isStarted))
      await _dataService.start();

    await _gpsService.init();

    if (!_detectionService.isStarted)
      _detectionService.start(NearbyConnectionsProximityDetection()).then((result) {
        _asynchronousSetUpFunctionFinished();
      });

    if (!_backgroundFetchService.isStarted)
      _backgroundFetchService.start().then((result) {
        _asynchronousSetUpFunctionFinished();
      });

    print("HEROES_FLUTTER_PLUGIN :: Started.");
    Log.addToLog("Heroes started.");

    _isStarted = true;
    return true;
  }

  /// Stops detection of other devices and its handling.
  /// Returns whether or not it was successful.
  Future<bool> stop() async
  {
    if (_isStopping)
      return true;

    if (!_isStarted || _isStarting)
      return false;

    _isStopping = true;
    _dataService.finishAllEncounters();
    _detectionService.stop().then((result) {
      _asynchronousStopFunctionFinished();
    });
    await _dataService.performBackUp();
    _isStarted = false;
    return true;
  }

  /// Performs backup of the local database of encounters.
  /// Returns whether or not it was successful.
  Future<bool> performBackUp() async
  {
    return await _dataService.performBackUp();
  }

  /// Returns [ExposureData]: the total duration of this device's encounters with infected devices and the timestamp of when this info was calculated.
  ExposureData getExposureData()
  {
    return _dataService.getTotalEncounterTime();
  }

  /// Sets log function callback to aid in development.
  /// The function must take a [String] parameter.
  void setLogCallback(Function function)
  {
    Log.setAddToLogCallback(function);
  }

  void _asynchronousSetUpFunctionFinished()
  {
    if (_detectionService.isStarted && _backgroundFetchService.isStarted) {
      _isStarting = false;
    }
  }

  void _asynchronousStopFunctionFinished()
  {
    if (!_dataService.isStarted) {
      _isStopping = false;
    }
  }

  @override
  String toString() => "Heroes(onExposureDetected: " +
      (onExposureDetected == null ? "null" : "not null") + ")";
}